from __future__ import annotations as _annotations

import ast
import asyncio
import dataclasses
import importlib.util
import inspect
import re
import sys
from collections.abc import Sequence
from dataclasses import dataclass
from importlib.abc import Loader
from pathlib import Path
from textwrap import indent
from typing import TYPE_CHECKING, Any, Callable
from unittest.mock import patch

import pytest
from black.parsing import InvalidInput

from .lint import black_format, code_diff
from .traceback import create_example_traceback

if TYPE_CHECKING:
    from .config import ExamplesConfig
    from .find_examples import CodeExample

__all__ = 'run_code', 'InsertPrintStatements', 'IncludePrint'

parent_frame_id = 4 if sys.version_info >= (3, 8) else 3
IncludePrint = Callable[[Path, inspect.FrameInfo, Sequence[Any]], bool]


def run_code(
    *,
    example: CodeExample,
    python_file: Path,
    loader: Loader | None,
    config: ExamplesConfig,
    enable_print_mock: bool,
    print_callback: Callable[[str], str] | None,
    include_print: IncludePrint | None,
    module_globals: dict[str, Any] | None,
    call: str | None,
) -> tuple[InsertPrintStatements, dict[str, Any]]:
    """Run the code example.

    Args:
        example: The `CodeExample` to run.
        python_file: The path to the python file.
        loader: optional loader to use to load the module.
        config: The `ExamplesConfig` to use.
        enable_print_mock: If True, mock the `print` function.
        print_callback: If not None, a callback to call on `print`.
        include_print: If not None, a function to call to determine if the print statement should be included.
        module_globals: The extra globals to add before calling the module.
        call: If not None, a (coroutine) function to call in the module.

    Returns:
        A tuple of the `InsertPrintStatements` instance and the module's globals.
    """
    __tracebackhide__ = True

    spec = importlib.util.spec_from_file_location('__main__', str(python_file), loader=loader)
    assert spec is not None, f'Could not load {python_file}'
    assert spec.loader is not None, f'Loader is None for {python_file}'
    module = importlib.util.module_from_spec(spec)

    # does nothing if insert_print_statements is False
    insert_print = InsertPrintStatements(python_file, config, enable_print_mock, print_callback, include_print)

    if module_globals:
        module.__dict__.update(module_globals)

    try:
        with insert_print:
            sys.modules[spec.name] = module
            spec.loader.exec_module(module)
            if call:
                to_call = getattr(module, call, None)
                if to_call is not None:
                    if inspect.iscoroutinefunction(to_call):
                        asyncio.run(to_call())
                    else:
                        to_call()
    except KeyboardInterrupt:
        print('KeyboardInterrupt in example')
    except Exception as exc:
        example_tb = create_example_traceback(exc, str(python_file), example)
        if example_tb:
            raise exc.with_traceback(example_tb)
        else:
            raise exc

    return insert_print, {k: v for k, v in module.__dict__.items() if not k.startswith(('__', '@'))}


@dataclass(init=False)
class Arg:
    """A single argument to a print statement."""

    data: str
    is_str: bool = False

    def __init__(self, v: Any):
        if isinstance(v, str):
            self.data = v
            self.is_str = True
        elif isinstance(v, set):
            # NOTE! this is not recursive
            ordered = ', '.join(repr(x) for x in sorted(v))
            self.data = f'{{{ordered}}}'
        else:
            self.data = re.sub('0x[a-f0-9]{8,12}>', '0x0123456789ab>', str(v))

    def __str__(self) -> str:
        return self.data

    def format(self, config: ExamplesConfig) -> str:
        if self.is_str:
            return self.data
        else:
            try:
                return black_format(self.data, config)
            except InvalidInput:
                return self.data


@dataclass
class PrintStatement:
    """A single print statement."""

    line_no: int
    sep: str
    args: list[Arg]

    def __str__(self):
        return self.sep.join(map(str, self.args))


def not_print(*args):
    import sys

    sys.stdout.write(' '.join(map(str, args)) + '\n')


class MockPrintFunction:
    __slots__ = 'file', 'statements', 'include_print'

    def __init__(self, file: Path, include_print: IncludePrint | None) -> None:
        self.file = file
        self.statements: list[PrintStatement] = []
        self.include_print = include_print

    def __call__(self, *args: Any, sep: str = ' ', **kwargs: Any) -> None:
        frame = inspect.stack()[parent_frame_id]

        if self._include_file(frame, args):
            lineno = self._find_line_number(frame)
            s = PrintStatement(lineno, sep, [Arg(arg) for arg in args])
            self.statements.append(s)

    def _include_file(self, frame: inspect.FrameInfo, args: Sequence[Any]) -> bool:
        if self.include_print:
            return self.include_print(self.file, frame, args)
        else:
            return self.file.samefile(frame.filename)

    def _find_line_number(self, inspect_frame: inspect.FrameInfo) -> int:
        """Find the line number of the print statement in the file that is being executed."""
        frame = inspect_frame.frame
        while True:
            if self.file.samefile(frame.f_code.co_filename):
                return frame.f_lineno
            elif frame.f_back:
                frame = frame.f_back
            else:
                raise RuntimeError(f'Could not find line number of print statement at {inspect_frame}')


class InsertPrintStatements:
    def __init__(
        self,
        python_path: Path,
        config: ExamplesConfig,
        enable: bool,
        print_callback: Callable[[str], str] | None,
        include_print: IncludePrint | None,
    ):
        self.file = python_path
        self.config = config
        self.print_func = MockPrintFunction(python_path, include_print) if enable else None
        self.print_callback = print_callback
        self.patch = None

    def __enter__(self) -> None:
        if self.print_func:
            self.patch = patch('builtins.print', side_effect=self.print_func)
            self.patch.start()

    def __exit__(self, *args) -> None:
        if self.patch is not None:
            self.patch.stop()

    def check_print_statements(self, example: CodeExample) -> None:
        new_code = self.updated_print_statements(example)
        if new_code is not None:
            diff = code_diff(example, new_code, self.config)
            pytest.fail(f'Print output changed code:\n{indent(diff, "  ")}', pytrace=False)

    def updated_print_statements(self, example: CodeExample) -> str | None:
        with_prints = self._insert_print_statements(example)
        # we check against the raw `with_prints` and `with_prints` with trailing whitespace removed
        # since trailing white space will have already been stripped by pre-commit in `example.source`
        if example.source not in (with_prints, re.sub(r'[ \t]+\n', '\n', with_prints)):
            return with_prints

    def print_statements(self) -> list[PrintStatement]:
        return self.print_func.statements if self.print_func else []

    def _insert_print_statements(self, example: CodeExample) -> str:
        assert self.print_func is not None, 'print statements not being inserted'

        lines = example.source.splitlines()

        old_line_no = -1

        for s in reversed(self.print_func.statements):
            line_no, col = find_print_location(example, s.line_no)

            # switch from 1-indexed line number to 0-indexed indexes into lines
            line_index = line_no - 1

            if s.line_no != old_line_no:
                remove_old_print(lines, line_index)
            self._insert_print_args(lines, s, example.in_py_file(), line_index, col)
            old_line_no = s.line_no

        return '\n'.join(lines) + '\n'

    def _insert_print_args(
        self, lines: list[str], statement: PrintStatement, in_python: bool, line_index: int, col: int
    ) -> None:
        formatted_args = [self.print_callback(str(arg)) if self.print_callback else str(arg) for arg in statement.args]
        single_line = statement.sep.join(formatted_args)
        indent_str = ' ' * col
        max_single_length = self.config.line_length - len(indent_str)
        if '\n' not in single_line and len(single_line) + len(comment_prefix) < max_single_length:
            lines.insert(line_index + 1, f'{indent_str}{comment_prefix}{single_line}')
        else:
            # if the statement is too long to go on one line, print each arg on its own line formatted with black
            sep = f'{statement.sep}\n'
            indent_config = dataclasses.replace(self.config, line_length=max_single_length)
            output = sep.join(arg.format(indent_config).strip('\n') for arg in statement.args)
            if self.print_callback:
                output = self.print_callback(output)
            # remove trailing whitespace
            output = re.sub(r' +$', '', output, flags=re.MULTILINE)
            # have to use triple single quotes in python since we're already in a double quotes docstring
            quote = "'''" if in_python else '"""'
            lines.insert(line_index + 1, indent(f'{quote}\n{output}\n{quote}', indent_str))


comment_prefix = '#> '
comment_prefix_re = re.compile(f'^ *{re.escape(comment_prefix)}', re.MULTILINE)
triple_quotes_prefix_re = re.compile('^ *(?:"{3}|\'{3})', re.MULTILINE)


def remove_old_print(lines: list[str], line_index: int) -> None:
    """Remove the old print statement."""
    try:
        next_line = lines[line_index + 1]
    except IndexError:
        # end of file
        return

    if triple_quotes_prefix_re.search(next_line):
        for i in range(2, 10_000):
            if triple_quotes_prefix_re.search(lines[line_index + i]):
                del lines[line_index + 1 : line_index + i + 1]
                return
        raise ValueError('Could not find end of triple quotes')
    else:
        try:
            while comment_prefix_re.search(lines[line_index + 1]):
                del lines[line_index + 1]
        except IndexError:
            # end of file
            pass


def find_print_location(example: CodeExample, line_no: int) -> tuple[int, int]:
    """Find the line and column of the print statement.

    Args:
        example: the `CodeExample`
        line_no: The line number on which the print statement starts or approx

    Return: tuple if `(line, column)` of the print statement
    """
    m = ast.parse(example.source, filename=example.path.name)
    return find_print(m, line_no) or (line_no, 0)


# ast nodes that have a body
with_body = (
    ast.Module,
    ast.FunctionDef,
    ast.If,
    ast.Try,
    ast.ExceptHandler,
    ast.With,
    ast.For,
    ast.AsyncFor,
    ast.AsyncFunctionDef,
    ast.AsyncWith,
    ast.While,
    ast.ClassDef,
)


def find_print(node: Any, line: int) -> tuple[int, int] | None:
    if isinstance(node, with_body):
        found_loc = find_print_in_body(node.body, line)
        if found_loc is not None:
            return found_loc
        if isinstance(node, ast.If):
            found_loc = find_print_in_body(node.orelse, line)
            if found_loc is not None:
                return found_loc
        elif isinstance(node, ast.Try):
            for node in node.handlers:
                found_loc = find_print(node, line)
                if found_loc is not None:
                    return found_loc
    elif isinstance(node, ast.Expr):
        return find_print(node.value, line)
    elif isinstance(node, ast.Call):
        if isinstance(node.func, ast.Name) and node.func.id == 'print' and node.lineno == line:
            return expr_last_line(node), node.col_offset
        return find_print(node.func, line)


def find_print_in_body(body: list[ast.stmt], line: int) -> tuple[int, int] | None:
    for node in body:
        found_loc = find_print(node, line)
        if found_loc is not None:
            return found_loc


def expr_last_line(c: ast.expr) -> int:
    if isinstance(c, ast.Constant):
        return c.lineno
    if isinstance(c, ast.Call):
        if c.keywords:
            return maybe_plus_1(c, expr_last_line(c.keywords[-1].value))
        elif c.args:
            return maybe_plus_1(c, expr_last_line(c.args[-1]))
        else:
            return c.lineno
    elif isinstance(c, (ast.List, ast.Tuple, ast.Set)):
        if c.elts:
            return maybe_plus_1(c, expr_last_line(c.elts[-1]))
        else:
            return c.lineno
    elif isinstance(c, (ast.ListComp, ast.SetComp, ast.DictComp, ast.GeneratorExp)):
        gen = c.generators[-1]
        if gen.ifs:
            return maybe_plus_1(c, expr_last_line(gen.ifs[-1]))
        else:
            return maybe_plus_1(c, expr_last_line(gen.iter))
    else:
        return c.lineno


def maybe_plus_1(c: ast.expr, last_arg_line: int) -> int:
    if c.lineno == last_arg_line:
        # all args are on the same line, assume the basic `print(x, y)` case
        return c.lineno
    else:
        # args are on multiple lines, assume the following format an extra line with `)`
        return last_arg_line + 1



from __future__ import annotations as _annotations

import hashlib
import tempfile
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from black.const import DEFAULT_LINE_LENGTH
from black.mode import Mode as BlackMode, TargetVersion as BlackTargetVersion

if TYPE_CHECKING:
    from typing import Literal


__all__ = 'ExamplesConfig', 'DEFAULT_LINE_LENGTH'


@dataclass
class ExamplesConfig:
    line_length: int = DEFAULT_LINE_LENGTH
    quotes: Literal['single', 'double', 'either'] = 'either'
    magic_trailing_comma: bool = True
    target_version: Literal['py37', 'py38', 'py39', 'py310', 'py311'] = 'py37'
    upgrade: bool = False
    isort: bool = False
    ruff_line_length: int | None = None
    ruff_select: list[str] | None = None
    ruff_ignore: list[str] | None = None
    white_space_dot: bool = False
    """If True, replace spaces with `·` in example diffs."""

    def black_mode(self):
        return BlackMode(
            line_length=self.line_length,
            target_versions={BlackTargetVersion[self.target_version.upper()]} if self.target_version else set(),
            string_normalization=self.quotes == 'double',
            magic_trailing_comma=self.magic_trailing_comma,
        )

    def hash(self) -> str:
        # str(self) should be a good identifier of a specific config
        return hashlib.md5(str(self).encode()).hexdigest()

    def ruff_config(self) -> tuple[str, ...]:
        config_lines: list[str] = []
        select: list[str] = []
        ignore: list[str] = []
        args: list[str] = []

        # line length is enforced by black
        if self.ruff_line_length is None:
            # if not ruff line length, ignore E501 which is line length errors
            # by default, ruff sets the line length to 88
            ignore.append('E501')
        else:
            args.append(f'--line-length={self.ruff_line_length}')

        if self.ruff_select:
            select.extend(self.ruff_select)

        if self.quotes == 'single':
            # enforce single quotes using ruff, black will enforce double quotes
            select.append('Q')
            config_lines.append("flake8-quotes = {inline-quotes = 'single', multiline-quotes = 'double'}")

        if self.target_version:
            args.append(f'--target-version={self.target_version}')

        if self.upgrade:
            select.append('UP')
        if self.isort:
            select.append('I')

        if self.ruff_ignore:
            ignore.extend(self.ruff_ignore)

        if select:
            # use extend to not disable default select
            args.append(f'--extend-select={",".join(select)}')
        if ignore:
            args.append(f'--ignore={",".join(ignore)}')

        if config_lines:
            config_toml = '\n'.join(config_lines)
            config_file = Path(tempfile.gettempdir()) / 'pytest-examples-ruff-config' / self.hash() / 'ruff.toml'
            if not config_file.exists() or config_file.read_text() != config_toml:
                config_file.parent.mkdir(parents=True, exist_ok=True)
                config_file.write_text(config_toml)

            args.append(f'--config={config_file}')

        return tuple(args)



from __future__ import annotations as _annotations

from itertools import groupby
from pathlib import Path
from textwrap import indent
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .find_examples import CodeExample


def _modify_files(examples: list[CodeExample]) -> str:
    """Internal use only, update examples in place."""
    # The same example shouldn't appear more than once
    unique_examples: set[str] = set()
    for ex in examples:
        s = str(ex)
        if s in unique_examples:
            examples_str = '\n'.join(f'  {ex} (test: {ex.test_id})' for ex in examples)
            raise RuntimeError(f'Cannot update the same example in separate tests!\nexamples:\n{examples_str}')
        unique_examples.add(s)

    # same file should not appear in more than one group
    files: set[Path] = set()
    # order by line number descending so the earlier change doesn't mess up line numbers for later changes
    examples.sort(key=lambda x: (x.group, x.start_line), reverse=True)

    for _, g in groupby(examples, key=lambda x: x.group):
        new_files = {ex.path for ex in g}
        if new_files & files:
            raise RuntimeError('Cannot update the same file in separate groups!')
        files |= new_files

    msg = [f'pytest-examples: {len(examples)} examples to update in {len(files)} file(s)...']

    for path, g in groupby(examples, key=lambda ex: ex.path):
        content = path.read_text()
        count = 0
        for ex in g:
            example: CodeExample = ex
            new_source = example.source
            if example.indent:
                new_source = indent(new_source, ' ' * example.indent)
            content = content[: example.start_index] + new_source + content[example.end_index :]
            count += 1

        msg.append(f'  {path} {count} examples updated')
        path.write_text(content)

    return '\n'.join(msg)



from __future__ import annotations as _annotations

from collections.abc import Iterator
from importlib.metadata import version
from pathlib import Path

import pytest

from .eval_example import EvalExample
from .find_examples import CodeExample, find_examples

__version__ = version('pytest_examples')
__all__ = 'find_examples', 'CodeExample', 'EvalExample'


def pytest_addoption(parser) -> None:
    """Add options to the pytest command line."""
    group = parser.getgroup('examples')
    group.addoption(
        '--update-examples',
        action='store_true',
        help=(
            '[WARNING: this will allow files to be changed in place!] '
            'Update code examples to reflect print output and linting.'
        ),
    )
    group.addoption(
        '--update-examples-disable-summary',
        action='store_true',
        help='Disable the summary of updated examples at the end of the test run.',
    )


summary: str | None = None


@pytest.fixture(scope='session')
def _examples_to_update(pytestconfig: pytest.Config) -> Iterator[list[CodeExample]]:
    """Don't use this directly, it's just  used by."""
    global summary

    examples_to_update: list[CodeExample] = []
    yield examples_to_update
    if pytestconfig.getoption('update_examples') and examples_to_update:
        from .modify_files import _modify_files

        summary_ = _modify_files(examples_to_update)
        if not pytestconfig.getoption('update_examples_disable_summary'):
            summary = summary_


@pytest.fixture(name='eval_example')
def eval_example(tmp_path: Path, request: pytest.FixtureRequest, _examples_to_update) -> Iterator[EvalExample]:
    """Fixture to return a `EvalExample` instance for running and linting examples."""
    eval_ex = EvalExample(tmp_path=tmp_path, pytest_request=request)
    yield eval_ex
    if request.config.getoption('update_examples'):
        _examples_to_update.extend(eval_ex.to_update)


def pytest_terminal_summary() -> None:
    """Customise pytest to print the summary of updated examples at the end of the test run."""
    if summary:
        print(summary)



from __future__ import annotations as _annotations

import re
from subprocess import PIPE, Popen
from textwrap import indent
from typing import TYPE_CHECKING

from black import format_str as black_format_str
from black.output import diff as black_diff
from ruff.__main__ import find_ruff_bin

from .config import ExamplesConfig

if TYPE_CHECKING:
    from .find_examples import CodeExample

__all__ = 'ruff_check', 'ruff_format', 'black_check', 'black_format', 'code_diff', 'FormatError'


class FormatError(ValueError):
    pass


def ruff_format(
    example: CodeExample,
    config: ExamplesConfig,
    *,
    ignore_errors: bool = False,
) -> str:
    args = ('--fix',)
    if ignore_errors:
        args += ('--exit-zero',)
    try:
        return ruff_check(example, config, extra_ruff_args=args)
    except FormatError:
        # this is a workaround for https://github.com/charliermarsh/ruff/issues/3694#issuecomment-1483388856
        try:
            ruff_check(example, config)
        except FormatError as e2:
            raise e2 from None
        else:
            raise Exception('ruff failed in Fix mode but not in Check mode, please report this')


def ruff_check(
    example: CodeExample,
    config: ExamplesConfig,
    *,
    extra_ruff_args: tuple[str, ...] = (),
) -> str:
    ruff = find_ruff_bin()
    args = ruff, 'check', '-', *config.ruff_config(), *extra_ruff_args

    p = Popen(args, stdin=PIPE, stdout=PIPE, stderr=PIPE, universal_newlines=True)
    stdout, stderr = p.communicate(example.source, timeout=10)
    if p.returncode == 1 and stdout:

        def replace_offset(m: re.Match[str]):
            line_number = int(m.group(1))
            return f' --> {example.path}:{line_number + example.start_line}'

        output = re.sub(r'^ --> -:(\d+)', replace_offset, stdout, flags=re.M)
        raise FormatError(f'ruff failed:\n{indent(output, "  ")}')
    elif p.returncode != 0:
        raise RuntimeError(f'Error running ruff, return code {p.returncode}:\n{stderr or stdout}')
    else:
        return stdout


def black_format(source: str, config: ExamplesConfig, *, remove_double_blank: bool = False) -> str:
    # hack to avoid black complaining about our print output format
    before_black = re.sub(r'^( *#)> ', r'\1 > ', source, flags=re.M)
    after_black = black_format_str(before_black, mode=config.black_mode())
    # then revert it back
    after_black = re.sub(r'^( *#) > ', r'\1> ', after_black, flags=re.M)
    if remove_double_blank:
        after_black = re.sub(r'\n{3}', '\n\n', after_black)
    return after_black


def black_check(example: CodeExample, config: ExamplesConfig) -> None:
    after_black = black_format(example.source, config, remove_double_blank=example.in_py_file())
    if example.source != after_black:
        diff = code_diff(example, after_black, config)
        raise FormatError(f'black failed:\n{indent(diff, "  ")}')


def code_diff(example: CodeExample, after: str, config: ExamplesConfig) -> str:
    diff = black_diff(sub_space(example.source, config), sub_space(after, config), 'before', 'after')

    def replace_at_line(match: re.Match[str]) -> str:
        offset = re.sub(r'\d+', lambda m: str(int(m.group(0)) + example.start_line), match.group(2))
        return f'{match.group(1)}{offset}{match.group(3)}'

    return re.sub(r'^(@@\s*)(.*)(\s*@@)$', replace_at_line, diff, flags=re.M)


def sub_space(text: str, config: ExamplesConfig) -> str:
    if config.white_space_dot:
        return text.replace(' ', '·')
    else:
        return text



from __future__ import annotations as _annotations

import re
import shlex
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent
from uuid import UUID, uuid4

__all__ = 'CodeExample', 'find_examples'


@dataclass
class CodeExample:
    """Information about a Python code example."""

    source: str
    """The source code of the example, this is has any indent removed."""
    path: Path
    """The path to the file containing the example."""
    start_line: int
    """The line number of the first line of the example."""
    end_line: int
    """The line number of the last line of the example."""
    start_index: int
    """Index of the start of the example."""
    end_index: int
    """Index of the end of the example."""
    prefix: str
    """The prefix of the code block, e.g. `py`, can also contain rules for skipping some tests."""
    indent: int
    """The indentation of the example, number of spaces."""
    group: UUID | None = None
    """A unique identifier for the example group."""
    test_id: str | None = None
    """ID of the test this example was generated for."""

    @classmethod
    def create(
        cls,
        source: str,
        *,
        path: Path = Path('testing.md'),
        start_line: int = 0,
        end_line: int | None = None,
        start_index: int = 0,
        end_index: int | None = None,
        prefix: str = '',
        indent: int = 0,
    ):
        """Create a `CodeExample`, mostly for testing."""
        if end_line is None:
            end_line = start_line + source.count('\n')
        if end_index is None:
            end_index = start_index + len(source)
        return cls(
            source=source,
            path=path,
            start_line=start_line,
            end_line=end_line,
            start_index=start_index,
            end_index=end_index,
            prefix=prefix,
            indent=indent,
        )

    @property
    def module_name(self) -> str:
        """A suitable Python module name for testing the example."""
        return f'{self.path.stem}_{self.start_line}_{self.end_line}'

    def prefix_settings(self) -> dict[str, str]:
        """Key/value pairs from the prefix line.

        This works on the format `py foo="bar" spam="with space"`.
        """
        settings = {}
        for m in re.finditer(r'([^{\s]+?)=([\'"])(.+?)\2', self.prefix):
            settings[m.group(1)] = m.group(3)
        return settings

    def prefix_tags(self) -> set[str]:
        """Extract tags from the prefix, alternative logic to `prefix_settings`.

        This works on the format `py .foo .bar` or `{.py .foo .bar}`.
        """
        tags = shlex.split(self.prefix.strip(' {}'))
        tags = {p.lstrip('.') for p in tags if p}
        try:
            tags.remove('py')
        except KeyError:
            pass

        return tags

    def in_py_file(self) -> bool:
        """Whether the example is in a Python file."""
        return self.path.suffix == '.py'

    def __str__(self):
        try:
            path = self.path.relative_to(Path.cwd())
        except ValueError:
            path = self.path
        return f'{path}:{self.start_line}-{self.end_line}'


def find_examples(*paths: str | Path, skip: bool = False) -> Iterable[CodeExample]:
    """Find Python code examples in markdown files and python file docstrings.

    :param paths: Directories or files to search for examples in.
    :param skip: Whether to exit early and not search for examples, useful when running on windows where search fails.
    :return: A generator of `CodeExample` objects.
    """
    if skip:
        return

    for s in paths:
        path = Path(s)
        if path.is_file():
            sub_paths = [path]
        elif path.is_dir():
            sub_paths = path.glob('**/*')
        else:
            raise ValueError(f'Not a file or directory: {s!r}')

        for path in sub_paths:
            group = uuid4()
            if path.suffix == '.py':
                code = path.read_text('utf-8')
                for m_docstring in re.finditer(r'(^ *)(r?""")(.+?)"""', code, flags=re.M | re.S):
                    start_line = code[: m_docstring.start()].count('\n')
                    docstring = m_docstring.group(3)
                    index_offset = m_docstring.start() + len(m_docstring.group(1)) + len(m_docstring.group(2))
                    yield from _extract_code_chunks(
                        path, docstring, group, line_offset=start_line, index_offset=index_offset
                    )
            elif path.suffix == '.md':
                code = path.read_text('utf-8')
                yield from _extract_code_chunks(path, code, group)


def _extract_code_chunks(
    path: Path, text: str, group: UUID, *, line_offset: int = 0, index_offset: int = 0
) -> Iterable[CodeExample]:
    for m_code in re.finditer(r'(^ *```)( *)(.*?)\n(.+?)\1', text, flags=re.M | re.S):
        group1, group2, prefix, source = m_code.groups()
        prefix = prefix.lower()
        if prefix.startswith(('py', '{.py')):
            start_line = line_offset + text[: m_code.start()].count('\n') + 1
            source_dedent, indent = remove_indent(source)
            # 1 for the newline
            start_index = index_offset + m_code.start() + len(group1) + len(group2) + len(prefix) + 1
            yield CodeExample(
                source=source_dedent,
                path=path,
                start_line=start_line,
                end_line=start_line + source.count('\n') + 1,
                start_index=start_index,
                end_index=start_index + len(source),
                prefix=prefix,
                indent=indent,
                group=group,
            )


def remove_indent(text: str) -> tuple[str, int]:
    """Remove the given indent from each line of text, return the dedented text and the indent."""
    first_line_before = text[: text.strip('\n').find('\n')]
    text = dedent(text)
    first_line_after = text[: text.strip('\n').find('\n')]
    return text, len(first_line_before) - len(first_line_after)



from __future__ import annotations as _annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable

import pytest
from _pytest.assertion.rewrite import AssertionRewritingHook
from _pytest.outcomes import Failed as PytestFailed

from .config import DEFAULT_LINE_LENGTH, ExamplesConfig
from .lint import FormatError, black_check, black_format, ruff_check, ruff_format
from .run_code import IncludePrint, InsertPrintStatements, run_code

if TYPE_CHECKING:
    from typing import Literal

    from .find_examples import CodeExample

__all__ = ('EvalExample',)


class EvalExample:
    """Class to run and lint examples."""

    def __init__(self, *, tmp_path: Path, pytest_request: pytest.FixtureRequest):
        self.tmp_path = tmp_path
        self._pytest_config = pytest_request.config
        self._test_id = pytest_request.node.nodeid
        self.to_update: list[CodeExample] = []
        self.config: ExamplesConfig = ExamplesConfig()
        self.print_callback: Callable[[str], str] | None = None
        self.include_print: IncludePrint | None = None

    def set_config(
        self,
        *,
        line_length: int = DEFAULT_LINE_LENGTH,
        quotes: Literal['single', 'double', 'either'] = 'either',
        magic_trailing_comma: bool = True,
        target_version: Literal['py37', 'py38', 'py39', 'py310', 'py310'] = 'py37',
        upgrade: bool = False,
        isort: bool = False,
        ruff_line_length: int | None = None,
        ruff_select: list[str] | None = None,
        ruff_ignore: list[str] | None = None,
    ):
        """Set the config for lints.

        Args:
            line_length: The line length to use when wrapping print statements, defaults to 88.
            quotes: The quote to use, defaults to "either".
            magic_trailing_comma: If True, add a trailing comma to magic methods, defaults to True.
            target_version: The target version to use when upgrading code, defaults to "py37".
            upgrade: If True, upgrade the code to the target version, defaults to False.
            isort: If True, run ruff's isort extension on the code, defaults to False.
            ruff_line_length: In general, we disable line-length checks in ruff, to let black take care of them.
            ruff_select: Ruff rules to select
            ruff_ignore: Ruff rules to ignore
        """
        self.config = ExamplesConfig(
            line_length=line_length,
            quotes=quotes,
            magic_trailing_comma=magic_trailing_comma,
            target_version=target_version,
            upgrade=upgrade,
            isort=isort,
            ruff_line_length=ruff_line_length,
            ruff_select=ruff_select,
            ruff_ignore=ruff_ignore,
        )

    @property
    def update_examples(self) -> bool:
        return bool(self._pytest_config.getoption('update_examples'))

    def run(
        self,
        example: CodeExample,
        *,
        module_globals: dict[str, Any] | None = None,
        rewrite_assertions: bool = True,
        call: str | None = None,
    ) -> dict[str, Any]:
        """Run the example, print is not mocked and print statements are not checked.

        Args:
            example: The example to run.
            module_globals: The globals to use when running the example.
            rewrite_assertions: If True, rewrite assertions in the example using pytest's assertion rewriting.
            call: If not None, method to check for and call if it exists.
        """
        __tracebackhide__ = True
        example.test_id = self._test_id
        _, module_dict = self._run(example, None, module_globals, rewrite_assertions, call)
        return module_dict

    def run_print_check(
        self,
        example: CodeExample,
        *,
        module_globals: dict[str, Any] | None = None,
        rewrite_assertions: bool = True,
        call: str | None = None,
    ) -> dict[str, Any]:
        """Run the example and check print statements.

        Args:
            example: The example to run.
            module_globals: The globals to use when running the example.
            rewrite_assertions: If True, rewrite assertions in the example using pytest's assertion rewriting.
            call: If not None, method to check for and call if it exists.
        """
        __tracebackhide__ = True
        example.test_id = self._test_id
        insert_print, module_dict = self._run(example, 'check', module_globals, rewrite_assertions, call)
        insert_print.check_print_statements(example)
        return module_dict

    def run_print_update(
        self,
        example: CodeExample,
        *,
        module_globals: dict[str, Any] | None = None,
        rewrite_assertions: bool = True,
        call: str | None = None,
    ) -> dict[str, Any]:
        """Run the example and update print statements, requires `--update-examples`.

        Args:
            example: The example to run.
            module_globals: The globals to use when running the example.
            rewrite_assertions: If True, rewrite assertions in the example using pytest's assertion rewriting.
            call: If not None, method to check for and call if it exists.
        """
        __tracebackhide__ = True
        self._check_update(example)
        insert_print, module_dict = self._run(example, 'update', module_globals, rewrite_assertions, call)

        new_code = insert_print.updated_print_statements(example)
        if new_code:
            example.source = new_code
            self._mark_for_update(example)
        return module_dict

    def _run(
        self,
        example: CodeExample,
        insert_print_statements: Literal['check', 'update', None],
        module_globals: dict[str, Any] | None,
        rewrite_assertions: bool,
        call: str | None,
    ) -> tuple[InsertPrintStatements, dict[str, Any]]:
        __tracebackhide__ = True

        if rewrite_assertions:
            loader = AssertionRewritingHook(config=self._pytest_config)
            loader.mark_rewrite(example.module_name)
        else:
            loader = None

        if insert_print_statements == 'check':
            enable_print_mock = True
        elif insert_print_statements == 'update':
            enable_print_mock = True
        else:
            enable_print_mock = False

        python_file = self._write_file(example)
        return run_code(
            example=example,
            python_file=python_file,
            loader=loader,
            config=self.config,
            enable_print_mock=enable_print_mock,
            print_callback=self.print_callback,
            include_print=self.include_print,
            module_globals=module_globals,
            call=call,
        )

    def lint(self, example: CodeExample) -> None:
        """Lint the example with black and ruff.

        Args:
            example: The example to lint.
        """
        self.lint_black(example)
        self.lint_ruff(example)

    def lint_black(self, example: CodeExample) -> None:
        """Lint the example using black.

        Args:
            example: The example to lint.
        """
        example.test_id = self._test_id
        try:
            black_check(example, self.config)
        except FormatError as exc:
            raise PytestFailed(str(exc), pytrace=False) from None

    def lint_ruff(
        self,
        example: CodeExample,
    ) -> None:
        """Lint the example using ruff.

        Args:
            example: The example to lint.
        """
        example.test_id = self._test_id
        try:
            ruff_check(example, self.config)
        except FormatError as exc:
            raise PytestFailed(str(exc), pytrace=False) from None

    def format(self, example: CodeExample) -> None:
        """Format the example with black and ruff, requires `--update-examples`.

        Args:
            example: The example to format.
        """
        self.format_ruff(example)
        self.format_black(example)

    def format_black(self, example: CodeExample) -> None:
        """Format the example using black, requires `--update-examples`.

        Args:
            example: The example to lint.
        """
        self._check_update(example)

        new_content = black_format(example.source, self.config, remove_double_blank=example.in_py_file())
        if new_content != example.source:
            example.source = new_content
            self._mark_for_update(example)

    def format_ruff(
        self,
        example: CodeExample,
    ) -> None:
        """Format the example using ruff, requires `--update-examples`.

        Args:
            example: The example to lint.
        """
        self._check_update(example)

        try:
            new_content = ruff_format(example, self.config)
        except FormatError as exc:
            raise PytestFailed(str(exc), pytrace=False) from None
        else:
            if new_content != example.source:
                example.source = new_content
                self._mark_for_update(example)

    def _check_update(self, example: CodeExample) -> None:
        if not self.update_examples:
            raise RuntimeError('Cannot update examples without --update-examples')
        example.test_id = self._test_id

    def _mark_for_update(self, example: CodeExample) -> None:
        """Add the example to self.to_update IF it's not already there."""
        s = str(example)
        if not any(s == str(ex) for ex in self.to_update):
            self.to_update.append(example)

    def _write_file(self, example: CodeExample) -> Path:
        python_file = self.tmp_path / f'{example.module_name}.py'
        python_file.write_text(example.source)
        return python_file



from __future__ import annotations as _annotations

import sys
from types import CodeType, FrameType, TracebackType
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .find_examples import CodeExample


def create_example_traceback(exc: Exception, module_path: str, example: CodeExample) -> TracebackType | None:
    """Create a new traceback with the filename and line numbers altered to match `example`.

    This involves lots of horrible hacking, but (somewhat miraculously) seems to work.

    Frames outside the example are not included in the new traceback.
    """
    frames = []
    tb = exc.__traceback__
    while tb is not None:
        frame = tb.tb_frame
        if frame.f_code.co_filename == module_path:
            frames.append((create_custom_frame(frame, example), tb.tb_lasti, tb.tb_lineno + example.start_line))
        tb = tb.tb_next

    frames.reverse()
    new_tb = None
    for altered_frame, lasti, lineno in frames:
        new_tb = TracebackType(tb_next=new_tb, tb_frame=altered_frame, tb_lasti=lasti, tb_lineno=lineno)
    return new_tb


def create_custom_frame(frame: FrameType, example: CodeExample) -> FrameType:
    """Create a new frame that mostly matches `frame` but with filename and line number faked.

    Create a new frame that mostly matches `frame` but with a code object that has
    a filename from `example` and adjusted first line number
    so that pytest shows the correct code context in the traceback.

    Taken mostly from https://naleraphael.github.io/blog/posts/devlog_create_a_builtin_frame_object/
    With the CodeType creation inspired by https://stackoverflow.com/a/16123158/949890.
    """
    import ctypes

    P_SIZE = ctypes.sizeof(ctypes.c_void_p)
    IS_X64 = P_SIZE == 8

    P_MEM_TYPE = ctypes.POINTER(ctypes.c_ulong if IS_X64 else ctypes.c_uint)

    ctypes.pythonapi.PyFrame_New.argtypes = (
        P_MEM_TYPE,  # PyThreadState *tstate
        P_MEM_TYPE,  # PyCodeObject *code
        ctypes.py_object,  # PyObject *globals
        ctypes.py_object,  # PyObject *locals
    )
    ctypes.pythonapi.PyFrame_New.restype = ctypes.py_object  # PyFrameObject*

    ctypes.pythonapi.PyThreadState_Get.argtypes = None  # type: ignore
    ctypes.pythonapi.PyThreadState_Get.restype = P_MEM_TYPE

    f_code = frame.f_code
    if sys.version_info >= (3, 12):
        code = CodeType(
            f_code.co_argcount,
            f_code.co_posonlyargcount,
            f_code.co_kwonlyargcount,
            f_code.co_nlocals,
            f_code.co_stacksize,
            f_code.co_flags,
            f_code.co_code,
            f_code.co_consts,
            f_code.co_names,
            f_code.co_varnames,
            str(example.path),
            f_code.co_name,
            f_code.co_qualname,
            f_code.co_firstlineno + example.start_line,
            f_code.co_linetable,
            f_code.co_exceptiontable,
        )
    elif sys.version_info >= (3, 11):
        code = CodeType(
            f_code.co_argcount,
            f_code.co_posonlyargcount,
            f_code.co_kwonlyargcount,
            f_code.co_nlocals,
            f_code.co_stacksize,
            f_code.co_flags,
            f_code.co_code,
            f_code.co_consts,
            f_code.co_names,
            f_code.co_varnames,
            str(example.path),
            f_code.co_name,
            f_code.co_qualname,
            f_code.co_firstlineno + example.start_line,
            f_code.co_lnotab,
            f_code.co_exceptiontable,
        )
    else:
        code = CodeType(
            f_code.co_argcount,
            f_code.co_posonlyargcount,
            f_code.co_kwonlyargcount,
            f_code.co_nlocals,
            f_code.co_stacksize,
            f_code.co_flags,
            f_code.co_code,
            f_code.co_consts,
            f_code.co_names,
            f_code.co_varnames,
            str(example.path),
            f_code.co_name,
            f_code.co_firstlineno + example.start_line,
            f_code.co_lnotab,
        )

    return ctypes.pythonapi.PyFrame_New(
        ctypes.pythonapi.PyThreadState_Get(),  # thread state
        ctypes.cast(id(code), P_MEM_TYPE),  # a code object
        frame.f_globals,  # a dict of globals
        frame.f_locals,  # a dict of locals
    )
